import re


class SimpleParser(object):
    def __init__(self, input, flags=0):
        self.buffer = input
        self.flags = flags

    def peek(self, regexp):
        p = re.compile(regexp, self.flags)
        m = p.match(self.buffer)
        if m: 
            token = m.group()
        else:
            token = ''
        return token

    def consume(self, regexp):
        p = re.compile(regexp, self.flags)
        m = p.match(self.buffer)
        if m: 
            token = m.group()
            self.buffer = self.buffer[len(token):]
        else:
            raise Exception("Unable to parse token: %s" % self.buffer[:10])
        return token

    def tokenize(self, regexps):
        while self:
            for regexp in regexps:
                try:
                    token = self.consume(regexp)
                    yield token
                    break
                except:
                    pass
            else:
                raise Exception("Unable to parse token: %s" % self.buffer[:10])

    def __nonzero__(self):
        return len(self.buffer)



from pydap.model import DatasetType
from pydap.lib import __dap__


class BaseResponse(object):
    def __init__(self, dataset):
        self.dataset = dataset
        self.headers = [
                ('XDODS-Server', 'dods/%s' % '.'.join([str(i) for i in __dap__])),
                ]

    @staticmethod
    def serialize(dataset):
        raise NotImplementedError(
                'Subclasses must implement serialize')

    def __call__(self, environ, start_response):
        start_response('200 OK', self.headers)

        if environ.get('x-wsgiorg.want_parsed_response'):
            return ResponseSerializer(self.dataset, self.serialize)
        else:
            return self.serialize(self.dataset)
        

class ResponseSerializer(object):
    """
    A serializer for responses that keeps the dataset
    for modification by WSGI middleware.

    Follows this specification::

      http://wsgi.org/wsgi/Specifications/avoiding_serialization

    """
    def __init__(self, dataset, serializer):
        self.dataset = dataset
        self.serializer = serializer

    def x_wsgiorg_parsed_response(self, type):
        if type is DatasetType:
            return self.dataset

    def __iter__(self):
        return self.serializer(self.dataset)



import traceback
from StringIO import StringIO

from pydap.lib import __dap__, INDENT, escape


class ErrorResponse(object):
    def __init__(self, dataset=None, info=None):
        self.info = info 

    def __call__(self, environ, start_response):
        headers = [('Content-description', 'dods_error'),
                   ('XDODS-Server', 'dods/%s' % '.'.join(str(d) for d in __dap__)),
                   ('Content-type', 'text/plain')]

        type_, value, traceback_ = self.info
        f = StringIO()
        traceback.print_exception(type_, value, traceback_, file=f)
        msg = f.getvalue()

        output = ['Error {\n',
                  INDENT, 'code = %s;\n' % getattr(type_, 'code', -1),
                  INDENT, 'message = %s;\n' % escape(msg),
                  '}']

        start_response('500 Internal Error', headers, self.info)
        return output



"""
HTML DAP response.

This is a simple HTML response, building a form to download data
in ASCII format. The response builds the HTML page and redirects
the user to the proper response (ASCII, usually) when a POST is
done.

"""

__author__ = "Roberto De Almeida <rob@pydap.org>"

import re
from urllib import quote

from pkg_resources import iter_entry_points
from paste.request import construct_url, parse_formvars
from paste.httpexceptions import HTTPSeeOther, HTTPBadRequest

from pydap.lib import __version__
from pydap.util.template import GenshiRenderer, StringLoader
from pydap.responses.dds import dispatch as dds_dispatch
from pydap.responses.lib import BaseResponse


COMMON_RESPONSES = ['dds', 'das', 'dods', 'asc', 'ascii', 'html', 'ver', 'version', 'help']


DEFAULT_TEMPLATE = """<html xmlns="http://www.w3.org/1999/xhtml"
      xmlns:py="http://genshi.edgewall.org/">
    <py:def function="attributes(attrs)">
        <dl py:if="attrs" class="attributes">
            <py:for each="k, v in attrs.items()">
                <dt>${k}</dt>
                <dd py:choose="">
                    <py:when test="isinstance(v, dict)">${attributes(v)}</py:when>
                    <py:otherwise>${v}</py:otherwise>
                </dd>
            </py:for>
        </dl>
    </py:def>

    <div py:def="children(var)" class="children">
        <fieldset py:for="child in var.walk()">
            <legend>${child.name}</legend>

            <?python
                from pydap.model import *
            ?>
            <py:choose test="">
                <p py:when="isinstance(child, SequenceType)" class="filter">
                    <select class="var1" name="var1_${child.id}[0]" id="var1_${child.id}[0]">
                        <option value="--" selected="selected">--</option>
                        <option py:for="grandchild in child.walk()" value="${grandchild.id}">${grandchild.id}</option>
                    </select>
                    <select class="op" name="op_${child.id}[0]" id="op_${child.id}[0]">
                        <option value="%3D" selected="selected">=</option>
                        <option value="%21%3D">!=</option>
                        <option value="%3C">&lt;</option>
                        <option value="%3C%3D">&lt;=</option>
                        <option value="%3E">&gt;</option>
                        <option value="%3E%3D">&gt;=</option>
                        <option value="%3D%7E">=~</option>
                    </select>
                    <input type="text" name="var2_${child.id}[0]" id="var2_${child.id}[0]" value="" class="var2" />
                </p>
                <div py:when="isinstance(child, (BaseType, GridType))">
                    <p><input type="checkbox" name="${child.id}" id="${child.id}" /><label for="${child.id}">Retrieve this variable at indices:</label></p>
                    <?python 
                        dimensions = child.dimensions or ['dim_%id' % j for j in range(len(child.shape))]
                    ?>
                    <div py:for="i in range(len(child.shape))">
                        <label for="${child.id}[${i}]">${dimensions[i]}</label>:<br />
                        <input type="text" name="${child.id}[${i}]" id="${child.id}[${i}]" value="0:1:${child.shape[i]-1}" /><br />
                    </div>
                </div>
            </py:choose>

            ${attributes(child.attributes)}
            ${children(child)}
        </fieldset>
    </div>

    <head>
        <title>Pydap server: ${dataset.name}</title>

        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    </head>

    <body>
        <div id="main">
            <h1>Download data from <code>${dataset.name}</code></h1>

            <!-- Global attributes -->
            <hr />
            <h2>Global attributes</h2>
            ${attributes(dataset.attributes)}

            <hr />
            <form id="dods_form" method="post" action="${location}">
                <!-- Dataset variables -->
                ${children(dataset)}

                <p><input type="submit" id="submit" value="Download data" /> as 
                <select id="response" name="response">
                    <option value="ascii" selected="selected">ASCII</option>
                    <option py:for="response in responses" value="$response">$response</option>
                </select>
                <br /><input type="reset" value="Reset" /></p>
            </form>

            <hr />
            <p><em><a href="http://pydap.org/">pydap/$version</a></em> &copy; Roberto De Almeida</p>
        </div>
    </body>
</html>"""


class HTMLResponse(BaseResponse):

    renderer = GenshiRenderer(
            options={}, loader=StringLoader( {'html.html': DEFAULT_TEMPLATE} ))

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend([
                ('Content-description', 'dods_form'),
                ('Content-type', 'text/html'),
                ])

    def __call__(self, environ, start_response):
        # If we came from a post, parse response and redirect to ascii.
        if environ['REQUEST_METHOD'] == 'POST':
            # Parse POST and redirect to proper response.
            form = parse_formvars(environ)
            projection, selection = [], []
            for k in form:
                # Selection.
                if k.startswith('var1') and form[k] != '--':
                    name = k[5:]
                    sel = '%s%s%s' % (form[k], form['op_%s' % name], form['var2_%s' % name])
                    selection.append(sel)
                # Projection.
                if form[k] == 'on':
                    var = [k]
                    for dim in form:
                        if re.match('%s\[\d+\]' % re.escape(k), dim):
                            var.append('[%s]' % form[dim])
                    var = ''.join(var)
                    if var not in projection:
                        projection.append(var)

            projection = ','.join(projection)
            selection = '&'.join(selection)
            query = projection + '&' + selection
            
            # Get current location.
            location = construct_url(environ, with_query_string=False)

            # Empty queries SHOULD NOT return everything, because this
            # means the user didn't select any variables.
            if not query.rstrip('?&'):
                app = HTTPBadRequest('You did not select any variables.')
            else:
                # Replace html extension for ascii and redirect.
                response = form['response']
                redirect = '%s.%s?%s' % (location[:-5], response, query)
                app = HTTPSeeOther(redirect)
            return app(environ, start_response)
        else:
            self.serialize = self._render(environ)
            return BaseResponse.__call__(self, environ, start_response)

    def _render(self, environ):
        def serialize(dataset):
            location = construct_url(environ, with_query_string=False)
            base = location[:location.rfind('/')]
            context = {
                'root': construct_url(environ, with_query_string=False, script_name='').rstrip('/'),
                'base': base,
                'location': location,
                'dataset': dataset,
                'responses': [r.name for r in iter_entry_points('pydap.response')
                        if r.name not in COMMON_RESPONSES],
                'version': '.'.join(str(i) for i in __version__),
            }
            renderer = environ.get('pydap.renderer', self.renderer)
            template = renderer.loader('html.html')
            output = renderer.render(template, context)
            if hasattr(dataset, 'close'): dataset.close()
            return [output]
        return serialize



import string

from paver.easy import *
from paver.setuputils import setup, find_packages, find_package_data
import paver.doctools
import paver.virtual
from paver.release import setup_meta

try:
    from pydap.lib import __version__
except ImportError:
    __version__ = ('unknown',)

options = environment.options
setup(**setup_meta)

options(
    setup=Bunch(
        name='Pydap',
        version='.'.join(str(d) for d in __version__),
        description='Pure Python Opendap/DODS client and server.',
        long_description='''
Pydap is an implementation of the Opendap/DODS protocol, written from
scratch. You can use Pydap to access scientific data on the internet
without having to download it; instead, you work with special array
and iterable objects that download data on-the-fly as necessary, saving
bandwidth and time. The module also comes with a robust-but-lightweight
Opendap server, implemented as a WSGI application.
        ''',
        keywords='opendap dods dap data science climate oceanography meteorology',
        classifiers=filter(None, map(string.strip, '''
            Development Status :: 5 - Production/Stable
            Environment :: Console
            Environment :: Web Environment
            Framework :: Paste
            Intended Audience :: Developers
            Intended Audience :: Science/Research
            License :: OSI Approved :: MIT License
            Operating System :: OS Independent
            Programming Language :: Python
            Topic :: Internet
            Topic :: Internet :: WWW/HTTP :: WSGI
            Topic :: Scientific/Engineering
            Topic :: Software Development :: Libraries :: Python Modules
        '''.split('\n'))),
        author='Roberto De Almeida',
        author_email='rob@pydap.org',
        url='http://pydap.org/',
        license='MIT',

        packages=find_packages(),
        package_data=find_package_data("pydap", package="pydap",
                only_in_packages=False),
        include_package_data=True,
        zip_safe=False,
        namespace_packages=['pydap', 'pydap.responses', 'pydap.handlers'],

        test_suite='nose.collector',

        dependency_links=[],
        install_requires=[
            'numpy',
            'httplib2>=0.4.0',
            'Genshi',
            'Paste',
            'PasteScript',
            'PasteDeploy',
        ],
        extras_require={
            'test': ['nose', 'wsgi_intercept'],
            'docs': ['Paver', 'Sphinx', 'Pygments', 'coards'],
        },
        entry_points="""
            [pydap.response]
            dds = pydap.responses.dds:DDSResponse
            das = pydap.responses.das:DASResponse
            dods = pydap.responses.dods:DODSResponse
            asc = pydap.responses.ascii:ASCIIResponse
            ascii = pydap.responses.ascii:ASCIIResponse
            ver = pydap.responses.version:VersionResponse
            version = pydap.responses.version:VersionResponse
            help = pydap.responses.help:HelpResponse
            html = pydap.responses.html:HTMLResponse
      
            [paste.app_factory]
            server = pydap.wsgi.file:make_app
      
            [paste.paster_create_template]
            pydap = pydap.wsgi.templates:DapServerTemplate
        """,
    ),
    minilib=Bunch(
        extra_files=['doctools', 'virtual']
    ), 
    virtualenv=Bunch(
        packages_to_install=['Paste', 'Pydap'],
        script_name='bootstrap.py',
        paver_command_line=None,
        install_paver=True
    ),
    sphinx=Bunch(
        builddir='_build',
    ),
    cog=Bunch(
        includedir='.',
    ),
    deploy=Bunch(
        htmldir = path('pydap.org'),
        username = 'roberto',
        host = 'pydap.org',
        hostpath = '/var/www/sites/pydap.org-3.x/'
    ),
)


if paver.doctools.has_sphinx:
    @task
    @needs(['cog', 'paver.doctools.html'])
    def html():
        """Build the docs and put them into our package."""
        destdir = path('pydap.org')
        destdir.rmtree()
        builtdocs = path("docs") / options.builddir / "html"
        builtdocs.move(destdir)

    @task
    @needs(['cog', 'paver.doctools.doctest'])
    def doctest():
        pass


if paver.virtual.has_virtualenv:
    @task
    def bootstrap():
        """Build a virtualenv bootstrap for developing paver."""
        paver.virtual._create_bootstrap(options.script_name,
                options.packages_to_install,
                options.paver_command_line,
                options.install_paver)


@task
@needs(['generate_setup', 'minilib', 'setuptools.command.sdist'])
def sdist():
    """Overrides sdist to make sure that our setup.py is generated."""
    pass


@task
@cmdopts([
    ('username=', 'u', 'Username to use when logging in to the servers')
])
def deploy():
    """Deploy the HTML to the server."""
    sh("rsync -avz -e ssh %s/ %s@%s:%s" % (options.htmldir,
            options.username, options.host, options.hostpath))




from paste.script import templates

class DapServerTemplate(templates.Template):

    egg_plugins = ['pydap']
    summary = 'A basic pydap server installation.'
    _template_dir = 'template'
    use_cheetah = False

    def post(self, command, output_dir, vars):
        if command.verbose:
            print '*'*72
            print '* Run "paster serve %s/server.ini" to run' % output_dir
            print '* the DAP server on http://localhost:8080/'
            print '*'*72



# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
except ImportError:
    # don't prevent use of pydap if pkg_resources isn't installed
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__) 

import modulefinder
for p in __path__:
    modulefinder.AddPackagePath(__name__, p)




"""
Ordered dictionary.

This is a dictionary class that preserves the order in which the keys are
stored. This is necessary to build Structures and Sequences that follow the
requested variable order.
"""

__author__ = "Roberto De Almeida <rob@pydap.org>"

import copy


class odict(dict):

    """Ordered dictionary."""
    
    def __init__(self, pairs=None, **kw):
        self._keys = []
        self._dict = {}

        if pairs is not None:
            for k, v in pairs:
                self[k] = v
        self.update(kw)

    def __iter__(self):
        return iter(self._keys[:])

    def __setitem__(self, key, item):
        self._dict.__setitem__(key, item)
        if key not in self._keys: self._keys.append(key)

    def __getitem__(self, key):
        return self._dict.__getitem__(key)

    def __delitem__(self, key):
        self._dict.__delitem__(key)
        self._keys.remove(key)

    def keys(self):
        return self._keys[:]

    def items(self):
        return [(key, self._dict.__getitem__(key)) for key in self._keys]

    def values(self):
        return [self._dict.__getitem__(key) for key in self._keys]

    def iterkeys(self):
        for key in self._keys: yield key

    def iteritems(self):
        for key in self._keys: yield (key, self._dict.__getitem__(key))

    def itervalues(self):
        for key in self._keys: yield self._dict.__getitem__(key)

    def clear(self):
        self._dict.clear()
        self._keys = []

    def copy(self):
        new = odict(self)
        return new

    def update(self, odict):
        for k, v in odict.items():
            self.__setitem__(k, v)

    def setdefault(self, key, d=None):
        if key not in self._keys: self._keys.append(key)
        return self._dict.setdefault(key, d)

    def get(self, key, d=None):
        if key in self._keys: return self._dict.__getitem__(key)
        else: return d

    def has_key(self, key):
        return self._dict.has_key(key)

    def popitem(self):
        try: key = self._keys[-1]
        except IndexError: raise KeyError('dictionary is empty')

        self._keys.remove(key)
        return self._dict.pop(key)

    def pop(self, key, d=None):
        value = self._dict.pop(key, d)

        try: self._keys.remove(key)
        except ValueError: pass
        
        return value

    def fromkeys(keys, d=None):
        new = odict()
        for key in keys: new.__setitem__(key, d)
        return new

    def __contains__(self, key):
        return self._dict.has_key(key)

    def __len__(self):
        return self._dict.__len__()

    def __repr__(self):
        return '{%s}' % ', '.join(['%s: %s' % (k.__repr__(), v.__repr__()) for (k, v) in self.items()])

    def __str__(self):
        return '{%s}' % ', '.join(['%s: %s' % (k.__repr__(), v.__repr__()) for (k, v) in self.items()])

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo=None, _nil=[]):
        new = odict()
        for k, v in self.items():
            new.__setitem__(k, copy.deepcopy(v))
        return new 


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()



# http://svn.pythonpaste.org/Paste/TemplateProposal/

import os


class TemplateNotFound(Exception):
    pass


class StringTemplateRenderer(object):
    name = 'string.Template'
    output_types = []

    def __init__(self, options, loader):
        self.options = options
        self.loader = loader

    def render(self, template_object, variables,
               fragment=False, output_format=None,
               output_type='unicode'):
        if output_type != 'unicode':
            raise ValueError(
                "Bad output_type: %r" % output_type)
        self.compile(template_object)
        return template_object.native.substitute(variables)

    def compile(self, template_object):
        import string
        s = template_object.unicode_contents()
        template_object.native = string.Template(s)


class MakoRenderer(object):
    name = 'mako'
    output_types = []

    def __init__(self, options, loader):
        self.options = options
        self.loader = loader

    def render(self, template_object, variables,
               fragment=False, output_format=None,
               output_type='unicode'):
        if output_type != 'unicode':
            raise ValueError(
                "Bad output_type: %r" % output_type)
        self.compile(template_object)
        return template_object.native.render(**variables)

    def compile(self, template_object):
        from mako.template import Template
        template_object.native = Template(filename=template_object.filename)


class GenshiRenderer(object):
    name = 'genshi'
    output_types = []

    def __init__(self, options, loader):
        self.options = options
        self.loader = loader

    def render(self, template_object, variables,
               fragment=False, output_format=None,
               output_type='unicode'):
        if output_type != 'unicode':
            raise ValueError(
                "Bad output_type: %r" % output_type)
        self.compile(template_object)
        stream = template_object.native.generate(**variables)
        return stream.render('html', doctype='html')

    def compile(self, template_object):
        from genshi.template import MarkupTemplate
        f = template_object.open()
        template_object.native = MarkupTemplate(f)


class FileSource(object):

    native = None
    # Ideally this would be picked up from the OS somehow:
    default_encoding = 'utf8'
    real_name = None

    def __init__(self, name, filename):
        self.name = name
        self.filename = filename
        self.real_name = filename
        self.mtime = None
        
    def open(self):
        if self.mtime is None:
            self.mtime = os.path.getmtime(self.filename)
        return open(self.filename, 'rb')

    def unicode_contents(self):
        return self.str_contents().decode(self.default_encoding)

    def str_contents(self):
        f = self.open()
        try:
            return f.read()
        finally:
            f.close()


class FileLoader(object):

    def __init__(self, base_directory):
        self.base_directory = base_directory

    def __call__(self, template_name, relative_to_template=None):
        if isinstance(template_name, FileSource):
            mtime = template_name.mtime
            if mtime is None:
                # It hasn't even been opened, so it must not be out of date
                return template_name
            cur_mtime = os.path.getmtime(template_name.filename)
            if mtime >= cur_mtime:
                return template_name
            # Not up-to-date
            template_name = template_name.name
            # template_name.name is an absolute name
            relative_to_template = None
        template_name = template_name.lstrip('/\\')
        template_name = template_name.replace(os.path.sep, '/')
        template_name = os.path.normpath(template_name)
        if relative_to_template is not None:
            template_name = os.path.join(
                os.path.dirname(relative_to_template), template_name)
        filename = os.path.join(self.base_directory, template_name)
        if not os.path.exists(filename):
            raise TemplateNotFound(
                "No template in %s (looked in %s)"
                % (template_name, filename))
        return FileSource(template_name, filename)


class StringSource(object):

    native = None
    default_encoding = 'utf-8'
    real_name = None

    def __init__(self, name, template):
        self.real_name = self.name = name
        self.template = template

    def open(self):
        from StringIO import StringIO
        return StringIO(self.template)

    def unicode_contents(self):
        return self.template.decode(self.default_encoding)

    def str_contents(self):
        return self.template


class StringLoader(dict):

    def __call__(self, template_name, relative_to_template=None):
        template = self[template_name]
        return StringSource(template_name, template)



import re
from urllib import unquote

from pydap.parsers import SimpleParser
from pydap.model import *

atomic_types = ('byte', 'int', 'uint', 'int16', 'uint16', 'int32', 
        'uint32', 'float32', 'float64', 'string', 'url')
constructors = ('grid', 'sequence', 'structure')


class DDSParser(SimpleParser):
    def __init__(self, dds):
        SimpleParser.__init__(self, dds, re.IGNORECASE)
        self.dds = dds

    def consume(self, regexp):
        token = SimpleParser.consume(self, regexp)
        self.buffer = self.buffer.lstrip()
        return token

    def _dataset(self):
        dataset = DatasetType()

        self.consume('dataset')
        self.consume('{')
        while self.peek('\w+').lower() in atomic_types + constructors:
            var = self._declaration()
            dataset[var.name] = var
        self.consume('}')

        dataset.name = unquote(self.consume('[^;]+'))
        dataset._set_id()
        self.consume(';')

        return dataset

    parse = _dataset

    def _declaration(self):
        token = self.peek('\w+')
        map = {'grid': self._grid,
               'sequence': self._sequence,
               'structure': self._structure}
        method = map.get(token.lower(), self._base_declaration)
        return method()

    def _base_declaration(self):
        type_ = self.consume('\w+')
        type_ = typemap[type_.lower()]
        name = unquote(self.consume('[^;\[]+'))

        shape, dimensions = self._dimensions()
        self.consume(';')

        var = BaseType(name=name, shape=shape,
                dimensions=dimensions, type=type_)
        return var

    def _dimensions(self):
        shape = []
        names = []
        while not self.peek(';'):
            self.consume('\[')
            token = self.consume('\w+')
            if self.peek('='):
                names.append(token)
                self.consume('=')
                token = self.consume('\d+')
            shape.append(int(token))
            self.consume('\]')
        return tuple(shape), tuple(names)

    def _sequence(self):
        sequence = SequenceType()
        self.consume('sequence')
        self.consume('{')

        while not self.peek('}'):
            var = self._declaration()
            sequence[var.name] = var
        self.consume('}')

        sequence.name = unquote(self.consume('[^;]+'))
        self.consume(';')
        return sequence

    def _structure(self):
        structure = StructureType()
        self.consume('structure')
        self.consume('{')

        while not self.peek('}'):
            var = self._declaration()
            structure[var.name] = var
        self.consume('}')

        structure.name = unquote(self.consume('[^;]+'))
        self.consume(';')
        return structure

    def _grid(self):
        grid = GridType()
        self.consume('grid')
        self.consume('{')

        self.consume('array')
        self.consume(':')
        array = self._base_declaration()
        grid[array.name] = array

        self.consume('maps')
        self.consume(':')
        while not self.peek('}'):
            var = self._base_declaration()
            grid[var.name] = var
        self.consume('}')

        grid.name = unquote(self.consume('[^;]+'))
        self.consume(';')
        return grid



import numpy 

from pydap.responses.dds import dispatch as dds_dispatch
from pydap.responses.lib import BaseResponse
from pydap.xdr import DapPacker
from pydap.lib import walk
from pydap.model import *


class DODSResponse(BaseResponse):
    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend([
                ('Content-description', 'dods_data'),
                ('Content-type', 'application/octet-stream'),
                ])

        size = calculate_size(dataset)
        if size is not None:
            self.headers.append(('Content-length', size))

    @staticmethod
    def serialize(dataset):
        # Generate DDS.
        for line in dds_dispatch(dataset):
            yield line
        yield 'Data:\n'
        for line in DapPacker(dataset):
            yield line
        if hasattr(dataset, 'close'): dataset.close()


def calculate_size(dataset):
    size = 0
    for var in walk(dataset):
        # Pydap can't calculate the size of a dataset with a 
        # Sequence since the data is streamed directly.
        if (isinstance(var, SequenceType) or
                (isinstance(var, BaseType) and
                    var.type in [Url, String])):
            return None
        elif isinstance(var, BaseType):
            # account for array size marker
            if var.shape:
                size += 8
            # calculate size
            if var.type == Byte:
                size += -numpy.prod(var.shape) % 4
            else:
                size += numpy.prod(var.shape) * var.type.size
    # account for DDS
    size += len(''.join(dds_dispatch(dataset))) + len('Data:\n')
    return size



from urllib import quote

from pydap.model import *
from pydap.model import typemap
from pydap.lib import INDENT, encode_atom, isiterable
from pydap.responses.lib import BaseResponse


class DASResponse(BaseResponse):
    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend([
                ('Content-description', 'dods_das'),
                ('Content-type', 'text/plain'),
                ])

    @staticmethod
    def serialize(dataset):
        output = ''.join(dispatch(dataset))
        if hasattr(dataset, 'close'): dataset.close()
        return [output]


def dispatch(var, level=0):
    dispatchers = [(DatasetType, _dataset),
                   (GridType, _base),
                   (StructureType, _structure),
                   (BaseType, _base)]

    for klass, func in dispatchers:
        if isinstance(var, klass):
            return func(var, level)


def _dataset(var, level=0):
    yield '%sAttributes {\n' % (level * INDENT)

    for attr, values in var.attributes.items():
        for line in _recursive_build(attr, values, level):
            yield line

    for child in var.walk():
        for line in dispatch(child, level=level+1):
            yield line
    yield '%s}\n' % (level * INDENT)


def _structure(var, level=0):
    yield '%s%s {\n' % (level * INDENT, quote(var.name))

    for attr, values in var.attributes.items():
        for line in _recursive_build(attr, values, level):
            yield line

    for child in var.walk():
        for line in dispatch(child, level=level+1):
            yield line
    yield '%s}\n' % (level * INDENT)


def _base(dapvar, level=0):
    yield '%s%s {\n' % (level * INDENT, quote(dapvar.name))

    for attr, values in dapvar.attributes.items():
        for line in _recursive_build(attr, values, level):
            yield line
    yield '%s}\n' % (level * INDENT)


def _recursive_build(attr, values, level=0):
    """
    Recursive function to build the DAS.
    
    """
    # Check for metadata.
    if isinstance(values, dict):
        yield '%s%s {\n' % ((level+1) * INDENT, attr)
        for k, v in values.items():
            for line in _recursive_build(k, v, level+1):
                yield line
        yield '%s}\n' % ((level+1) * INDENT)
    else:
        type_ = get_type(values).descriptor
        if not isiterable(values):
            values = [values]

        # Encode values
        values = [encode_atom(atom) for atom in values]
        yield '%s%s %s %s;\n' % ((level+1) * INDENT, type_,
                attr.replace(' ', '_'), ', '.join(values))


def get_type(values):
    # Direct conversion for arrays.
    if hasattr(values, 'dtype'):
        return typemap[values.dtype.char]  # numpy.array
    elif hasattr(values, 'typecode'):
        return typemap[values.typecode]    # array.array

    # Map Python type to DAP type.
    if not isiterable(values):
        values = [values]
    types = [typeconvert(atom) for atom in values]
    precedence = [String, Float64, Int32]
    types.sort(key=precedence.index)
    return types[0]


def typeconvert(obj):
    """Type conversion between Python and DODS, for the DAS."""
    types = [(basestring, String),
             (float, Float64),
             (long, Int32),
             (int, Int32)]
    for klass, type_ in types:
        if isinstance(obj, klass):
            return type_



from wsgi_intercept import add_wsgi_intercept
from wsgi_intercept import httplib2_intercept
httplib2_intercept.install()

from pydap.lib import isiterable
from pydap.handlers.lib import SimpleHandler


def UnitTestServer(dataset, host='localhost', port=8080, script_name=''):
    app = SimpleHandler(dataset)
    add_wsgi_intercept(host, port, lambda: app, script_name=script_name)


def to_list(L):
    if isiterable(L): return [to_list(item) for item in L]
    else: return L



"""
Client interface to OPeNDAP servers.

This module implements the ``open_url`` function, which returns an
object representing a remote dataset served by an OPeNDAP server. The
client builds the dataset representation from the DDS+DAS responses,
though in the future it can be extended to support other representations
(like DDX, or perhaps JSON).

"""

import sys
from urlparse import urlsplit, urlunsplit

from pydap.model import *
from pydap.model import DapType
from pydap.proxy import *
from pydap.parsers.dds import DDSParser
from pydap.parsers.das import DASParser
from pydap.xdr import DapUnpacker
from pydap.util.http import request
from pydap.exceptions import ClientError
from pydap.lib import walk, combine_slices, fix_slice, parse_qs, fix_shn, encode_atom


def open_url(url):
    """
    Open a given dataset URL, trying different response methods. 

    The function checks the stub DDX method, and falls back to the
    DDS+DAS responses. It can be easily extended for other representations
    like JSON.

    The URL should point to the dataset, omitting any response extensions
    like ``.dds``. Username and password can be passed in the URL like::

        http://user:password@example.com:port/path

    They will be transmitted as plaintext if the server supports only
    Basic authentication, so be careful. For Digest authentication this
    is safe.

    The URL can point directly to an Opendap dataset, or it can contain
    any number of contraint expressions (selection/projections)::

        http://example.com/dataset?var1,var2&var3>10

    You can also specify a cache directory, a timeout and a proxy using
    the global variables from ``pydap.lib``::

        >>> import pydap.lib
        >>> pydap.lib.TIMEOUT = 60  # seconds
        >>> pydap.lib.CACHE = '.cache'
        >>> import httplib2
        >>> from pydap.util import socks
        >>> pydap.lib.PROXY = httplib2.ProxyInfo(socks.PROXY_TYPE_HTTP, 'localhost', 8000)

    """
    for response in [_ddx, _ddsdas]:
        dataset = response(url)
        if dataset: break
    else:
        raise ClientError("Unable to open dataset.")

    # Remove any projections from the url, leaving selections.
    scheme, netloc, path, query, fragment = urlsplit(url)
    projection, selection = parse_qs(query)
    url = urlunsplit(
            (scheme, netloc, path, '&'.join(selection), fragment))

    # Set data to a Proxy object for BaseType and SequenceType. These
    # variables can then be sliced to retrieve the data on-the-fly.
    for var in walk(dataset, BaseType):
        var.data = ArrayProxy(var.id, url, var.shape)
    for var in walk(dataset, SequenceType):
        var.data = SequenceProxy(var.id, url)

    # Set server-side functions.
    dataset.functions = Functions(url)

    # Apply the corresponding slices.
    projection = fix_shn(projection, dataset)
    for var in projection:
        target = dataset
        while var:
            token, slice_ = var.pop(0)
            target = target[token]
            if slice_ and isinstance(target.data, VariableProxy):
                shape = getattr(target, 'shape', (sys.maxint,))
                target.data._slice = fix_slice(slice_, shape)

    return dataset


class Functions(object):
    """
    An entry point for server side functions.

    Since we don't know which functions the server support, any function
    can be called by the user. The name will be passed to the server, together
    with the corresponding arguments.

    """
    def __init__(self, baseurl):
        self.baseurl = baseurl

    def __getattr__(self, attr):
        return ServerFunction(self.baseurl, attr)


class ServerFunction(object):
    """
    An object representing a named server side function.

    The function uses lazy evaluation to allow for nested calls. Calling
    the function will return a ``ServerFunctionResult`` object that only
    downloads data when actually accessed by the user.

    """
    def __init__(self, baseurl, name):
        self.baseurl = baseurl
        self.name = name

    def __call__(self, *args):
        params = []
        for arg in args:
            if isinstance(arg, (DapType, ServerFunctionResult)):
                params.append(arg.id)
            else:
                params.append(encode_atom(arg))
        id_ = self.name + '(' + ','.join(params) + ')'
        return ServerFunctionResult(self.baseurl, id_)


class ServerFunctionResult(object):
    """
    An object containing the result from a server function call.

    The result will download data when any attribute or item is
    accessed, simulating a dataset object.

    """
    def __init__(self, baseurl, id_):
        self.id = id_
        self.dataset = None

        scheme, netloc, path, query, fragment = urlsplit(baseurl)
        self.url = urlunsplit((
                scheme, netloc, path + '.dods', id_, None))
    
    def __getattr__(self, name):
        if self.dataset is None: self.dataset = open_dods(self.url, True)
        return getattr(self.dataset, name)

    def __getitem__(self, key):
        if self.dataset is None: self.dataset = open_dods(self.url, True)
        return self.dataset[key]


def open_dods(url, get_metadata=False):
    """
    Download data from a DODS response and build a dataset.

    This function is useful to open "raw" URLs.

    """
    resp, data = request(url)
    dds, xdrdata = data.split('\nData:\n', 1)
    dataset = DDSParser(dds).parse()
    dataset.data = DapUnpacker(xdrdata, dataset).getvalue()

    if get_metadata:
        scheme, netloc, path, query, fragment = urlsplit(url)
        dasurl = urlunsplit(
                (scheme, netloc, path[:-5] + '.das', query, fragment))
        resp, das = request(dasurl)
        dataset = DASParser(das, dataset).parse()

    return dataset


def _ddx(url):
    """
    Stub function for DDX.

    Still waiting for the DDX spec to write this.

    """
    pass


def _ddsdas(url):
    """
    Build the dataset from the DDS+DAS responses.

    This function builds the dataset object from the DDS and DAS
    responses, adding Proxy objects to the variables.

    """
    scheme, netloc, path, query, fragment = urlsplit(url)
    ddsurl = urlunsplit(
            (scheme, netloc, path + '.dds', query, fragment))
    dasurl = urlunsplit(
            (scheme, netloc, path + '.das', query, fragment))

    respdds, dds = request(ddsurl)
    respdas, das = request(dasurl)

    # Build the dataset structure and attributes.
    dataset = DDSParser(dds).parse()
    dataset = DASParser(das, dataset).parse()
    return dataset




import sys
import re
from urlparse import urlsplit, urlunsplit
from urllib import quote
import copy
import warnings 

from pydap.model import *
from pydap.model import SequenceData
from pydap.lib import hyperslab, combine_slices, fix_slice, walk, isiterable
from pydap.util.http import request
from pydap.parsers.dds import DDSParser
from pydap.xdr import DapUnpacker


__all__ = ['VariableProxy', 'ArrayProxy', 'SequenceProxy']


class VariableProxy(object):
    """
    A proxy object pointing to an Opendap variable.

    The Opendap client works by defining proxy objects that point to
    specific variables in specific datasets. The object implements a
    typical interface (like an ndarray from Numpy) on the user-side,
    and downloads data on-the-fly from the server as necessary.

    This particular class is just an abstract implementation.

    """
    def __init__(self, id, url, slice_=None):
        self.id = id
        self.url = url
        self._slice = slice_ or (slice(None),)

    def __repr__(self):
        return '<%s pointing to variable "%s%s" at "%s">' % (
                self.__class__.__name__, self.id, hyperslab(self._slice), self.url)

    def __deepcopy__(self, memo=None, _nil=[]):
        out = self.__class__(self.id, self.url, self._slice)
        return out


class ConstraintExpression(object):
    """
    An object representing a given constraint expression.

    These objects are used to build the constraint expression used when
    downloading data. They are produced when a SequenceProxy is compared
    to something::

        >>> from pydap.model import *
        >>> s = SequenceType(name='s')
        >>> s['x'] = BaseType(name='x')
        >>> s.data = SequenceProxy('s', 'http://example.com/dataset')
        >>> s.x.data = SequenceProxy('s.x', 'http://example.com/dataset')
        >>> print type(s.x > 10)
        <class 'pydap.proxy.ConstraintExpression'>
        
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __and__(self, other):
        """Join two CEs together."""
        return self.__class__('%s&%s' % (self.value, other))

    def __or__(self, other):
        raise Exception('OR constraints not allowed in the Opendap specification.')


class ArrayProxy(VariableProxy):
    """
    Proxy to an Opendap basetype.

    """
    def __init__(self, id, url, shape, slice_=None):
        self.id = id
        self.url = url
        self.shape = shape

        if slice_ is None:
            self._slice = (slice(None),) * len(shape)
        else:
            self._slice = slice_ + (slice(None),)*(len(shape)-len(slice_))

    @property
    def __array_interface__(self):
        data = self[:]
        return {
                'version': 3,
                'shape': data.shape,
                'typestr': data.dtype.str,
                'data': data,
                }

    def __iter__(self):
        return iter(self[:])

    def __getitem__(self, index):
        slice_ = combine_slices(self._slice, fix_slice(index, self.shape))
        scheme, netloc, path, query, fragment = urlsplit(self.url)
        url = urlunsplit((
                scheme, netloc, path + '.dods',
                quote(self.id) + hyperslab(slice_) + '&' + query,
                fragment))

        resp, data = request(url)
        dds, xdrdata = data.split('\nData:\n', 1)
        dataset = DDSParser(dds).parse()
        data = data2 = DapUnpacker(xdrdata, dataset).getvalue()

        # Retrieve the data from any parent structure(s).
        for var in walk(dataset):
            if type(var) in (StructureType, DatasetType):
                data = data[0]
            elif var.id == self.id: 
                return data

        # Some old servers return the wrong response. :-/
        # I found a server that would return an array to a request
        # for an array inside a grid (instead of a structure with
        # the array); this will take care of it.
        for var in walk(dataset):
            if type(var) in (StructureType, DatasetType):
                data2 = data2[0]
            elif self.id.endswith(var.id):
                return data2
            
    # Comparisons return a boolean array
    def __eq__(self, other): return self[:] == other 
    def __ne__(self, other): return self[:] != other
    def __ge__(self, other): return self[:] >= other
    def __le__(self, other): return self[:] <= other
    def __gt__(self, other): return self[:] > other
    def __lt__(self, other): return self[:] < other


class SequenceProxy(VariableProxy, SequenceData):
    """
    Proxy to an Opendap Sequence.

    This class simulates the behavior of a Numpy record array, proxying
    the data in an Opendap Sequence object (or a child variable inside
    a Sequence)::

        >>> from pydap.model import *
        >>> s = SequenceType(name='s')
        >>> s['id'] = BaseType(name='id')
        >>> s['x'] = BaseType(name='x')
        >>> s['y'] = BaseType(name='y')
        >>> s.data = SequenceProxy('s', 'http://example.com/dataset')
        >>> print s.data
        <SequenceProxy pointing to variable "s" at "http://example.com/dataset">
        >>> print s.x.data
        <SequenceProxy pointing to variable "s.x" at "http://example.com/dataset">

    We can use the same methods we would use if the data were local::

        >>> print s[0].x.data
        <SequenceProxy pointing to variable "s[0:1:0].x" at "http://example.com/dataset">
        >>> print s[10:20][2].y.data
        <SequenceProxy pointing to variable "s[12:1:12].y" at "http://example.com/dataset">
        >>> print s[ (s['id'] > 1) & (s.x > 10) ].data
        <SequenceProxy pointing to variable "s" at "http://example.com/dataset?s.id>1&s.x>10&">
        >>> print s[ ('y', 'x') ].data
        <SequenceProxy pointing to variable "s.y,s.x" at "http://example.com/dataset">
        >>> s2 = s[ ('y', 'x') ]
        >>> print s2[ s2.x > 10 ].x.data
        <SequenceProxy pointing to variable "s.x" at "http://example.com/dataset?s.x>10&">
        >>> print s[ ('y', 'x') ][0].data
        <SequenceProxy pointing to variable "s.y,s.x[0:1:0]" at "http://example.com/dataset">

    (While the last line may look strange, it's equivalent to
    ``s.y[0:1:0],s.x[0:1:0]`` -- at least on Hyrax).

    """
    def __init__(self, id, url, slice_=None, children=None):
        VariableProxy.__init__(self, id, url, slice_)
        self.children = children or ()

    def __repr__(self):
        id_ = ','.join('%s.%s' % (self.id, child) for child in self.children) or self.id
        return '<%s pointing to variable "%s%s" at "%s">' % (
                self.__class__.__name__, id_, hyperslab(self._slice), self.url)

    def __iter__(self):
        scheme, netloc, path, query, fragment = urlsplit(self.url)
        id_ = ','.join('%s.%s' % (self.id, child) for child in self.children) or self.id
        url = urlunsplit((
                scheme, netloc, path + '.dods',
                quote(id_) + hyperslab(self._slice) + '&' + query,
                fragment))

        resp, data = request(url)
        dds, xdrdata = data.split('\nData:\n', 1)
        dataset = DDSParser(dds).parse()
        dataset.data = DapUnpacker(xdrdata, dataset).getvalue()
        dataset._set_id()

        # Strip any projections from the request id.
        id_ = re.sub('\[.*?\]', '', self.id)
        for var in walk(dataset):
            if var.id == id_:
                data = var.data
                if isinstance(var, SequenceType):
                    order = [var.keys().index(k) for k in self.children]
                    data = reorder(order, data, var._nesting_level)
                return iter(data)

    def __len__(self):
        return len(list(self.__iter__()))

    def __getitem__(self, key):
        out = copy.deepcopy(self)
        if isinstance(key, ConstraintExpression):
            scheme, netloc, path, query, fragment = urlsplit(self.url)
            out.url = urlunsplit((
                    scheme, netloc, path, str(key & query), fragment))

            if out._slice != (slice(None),):
                warnings.warn('Selection %s will be applied before projection "%s".' % (
                        key, hyperslab(out._slice)))
        elif isinstance(key, basestring):
            out._slice = (slice(None),)
            out.children = ()
            parent = self.id
            if ',' in parent:
                parent = parent.split(',', 1)[0].rsplit('.', 1)[0]
            out.id = '%s%s.%s' % (parent, hyperslab(self._slice), key)
        elif isinstance(key, tuple):
            out.children = key[:]
        else:
            out._slice = combine_slices(self._slice, fix_slice(key, (sys.maxint,)))
        return out

    def __deepcopy__(self, memo=None, _nil=[]):
        out = self.__class__(self.id, self.url, self._slice, self.children[:])
        return out

    # Comparisons return a ``ConstraintExpression`` object
    def __eq__(self, other): return ConstraintExpression('%s=%s' % (self.id, other))
    def __ne__(self, other): return ConstraintExpression('%s!=%s' % (self.id, other))
    def __ge__(self, other): return ConstraintExpression('%s>=%s' % (self.id, other))
    def __le__(self, other): return ConstraintExpression('%s<=%s' % (self.id, other))
    def __gt__(self, other): return ConstraintExpression('%s>%s' % (self.id, other))
    def __lt__(self, other): return ConstraintExpression('%s<%s' % (self.id, other))


def reorder(order, data, level):
    """
    Reorder Sequence data according to the request.

    """
    if not (order and isiterable(data)):
        return data
    if level == 0:
        return tuple(data[i] for i in order)
    else:
        return [reorder(order, value, level-1) for value in data]


def _test():
    import doctest
    doctest.testmod()


if __name__ == "__main__":
    _test()



import re
import array
import operator

import numpy

from pydap.parsers import SimpleParser
from pydap.model import *
from pydap.util.safeeval import expr_eval


atomic_types = ('byte', 'int', 'uint', 'int16', 'uint16', 'int32',
        'uint32', 'float32', 'float64', 'string', 'url', 'alias')


class DASParser(SimpleParser):
    def __init__(self, das, dataset):
        SimpleParser.__init__(self, das, re.IGNORECASE)
        self.das = das
        self.dataset = dataset

    def consume(self, regexp):
        token = SimpleParser.consume(self, regexp)
        self.buffer = self.buffer.lstrip()
        return token

    def parse(self):
        self._target = self.dataset

        self.consume('attributes')
        self.consume('{')
        while not self.peek('}'):
            self._attr_container()
        self.consume('}')
        
        return self.dataset

    def _attr_container(self):
        if self.peek('\w+').lower() in atomic_types:
            name, values = self._attribute()
            self._target.attributes[name] = values
        else:
            self._container()

    def _container(self):
        name = self.consume('[\w_\.%]+')
        self.consume('{')

        if '.' in name:
            # Get new target.
            names = name.split('.')
            target = self._target
            d = [target] + names
            try:
                self._target = reduce(operator.getitem, d)
            except KeyError:
                pass

            # Get attributes.
            while not self.peek('}'):
                self._attr_container()
            self.consume('}')

            # Revert to old target.
            self._target = target

        elif isinstance(self._target, StructureType) and name in self._target:
            target = self._target
            self._target = target[name]

            # Get attributes.
            while not self.peek('}'):
                self._attr_container()
            self.consume('}')

            # Revert to old target.
            self._target = target

        else:
            self._target.attributes[name] = self._metadata()
            self.consume('}')

    def _metadata(self):
        output = {}
        while not self.peek('}'):
            if self.peek('\w+').lower() in atomic_types:
                name, values = self._attribute()
                output[name] = values
            else:
                name = self.consume('\w+')
                self.consume('{')
                output[name] = self._metadata()
                self.consume('}')
        return output

    def _attribute(self):
        type_ = self.consume('\w+')
        name = self.consume('[\w-]+')

        values = []
        while not self.peek(';'):
            value = self.consume(r'".*?[^\\]"|[^;,]+')
            
            if type_.lower() in ['string', 'url']:
                ##value = expr_eval(value)
                value = expr_eval(repr(value))
                value = value.strip('"')
            elif type_.lower() == 'alias':
                # Support for Alias is not documented in the DAP spec. I based
                # this on the Java documentation from the OPeNDAP website at:
                # http://www.opendap.org/api/javaDocs/dods/dap/Alias.html

                # Check if we should start from the root dataset or from
                # the current item.
                if value.startswith('.'):
                    tokens = value[1:].split('.')
                    target = self.dataset
                else:
                    tokens = value.split('.')
                    target = self._target
                    
                # Run over tokens to get the value.
                for token in tokens:
                    if (isinstance(target, StructureType) and
                            token in target):
                        value = target = target[token]
                    else:
                        value = target = target.attributes[token]
            else:
                if value.lower() == 'nan':
                    value = numpy.NaN
                else:
                    value = expr_eval(value)

                # Convert to proper type. This is specially important for floats,
                # since this preserves the resolution even though Python has no
                # difference between float 32 vs. 64.
                dtype = {'float64': 'd',
                         'float32': 'f',
                         'int32'  : 'l',
                         'int16'  : 'h',
                         'uint32' : 'L',
                         'uint16' : 'H',
                         'byte'   : 'B'}[type_.lower()]
                try:
                    value = array.array(dtype, [value])[0]
                except OverflowError:
                    value = long(value)

            values.append(value)
            if self.peek(','):
                self.consume(',')

        self.consume(';')

        if len(values) == 1:
            values = values[0]

        return name, values



import sys
import re
import string
import urllib

import numpy

from pydap.exceptions import ConstraintExpressionError


__author__  = 'Roberto De Almeida <rob@pydap.org>'
__version__ = (3,0,'rc',4)       # pydap version
__dap__     = (2,0)             # protocol version


# Global variables -- I know we should try to avoid them, but we need
# a place to put module configuration. These two shouldn't normally
# be altered.
USER_AGENT = 'pydap/%s' % '.'.join(str(d) for d in __version__)
INDENT = ' ' * 4

# A directory where the client should store cache information/data.
CACHE = None

# Socket level timeout (in seconds).
TIMEOUT = None

# A ``ProxyInfo`` instance. To use a proxy @ http://localhost:8000/::
#
# >>> import httplib2, pydap.util.socks, pydap.lib
# >>> pydap.lib.PROXY = httplib2.ProxyInfo(pydap.util.socks.PROXY_TYPE_HTTP, 'localhost', 8000)
PROXY = None


def isiterable(obj):
    """
    Check if an object is iterable (strings don't count).

        >>> isiterable("string")
        False
        >>> isiterable(1)
        False
        >>> isiterable([1])
        True

    """
    if isinstance(obj, basestring): return False

    try:
        iter(obj)
        return True
    except TypeError:
        return False


def quote(name):
    """
    Extended quote for the DAP spec.

    The period MUST be escaped in names (DAP spec, item 5.1):

        >>> quote("White space")
        'White%20space'
        >>> urllib.quote("Period.")
        'Period.'
        >>> quote("Period.")
        'Period%2E'

    """
    def escape_(c):
        if c == '.':
            c = '%2E'
        elif c not in string.letters + string.digits + '''_!~*'-"''' + '%':
            c = urllib.quote(c)
        return c
    return ''.join(map(escape_, name))


def escape(s):
    r"""
    Escape a string, which may contain quotes.

        >>> print escape('test')
        "test"
        >>> print escape('this is a "simple" test')
        "this is a \"simple\" test"

    """
    s = s.replace(r'\\', r'\\\\')
    s = s.replace(r'"', r'\"')
    s = '"%s"' % s
    return s


def encode_atom(atom):
    """
    Encode an atom, according to the DAP spec.

        >>> print encode_atom(None)
        NaN
        >>> print encode_atom('string')
        "string"
        >>> print encode_atom(1)
        1
        >>> print encode_atom(1/3.)
        0.333333

    """
    if atom is None:
        return 'NaN'
    if isinstance(atom, basestring):
        return escape(atom)
    try:
        return '%.6g' % atom
    except TypeError:
        return escape(str(atom))


def walk(var, type_=object):
    """
    Yield all variables of a given type from a dataset.

    """
    if isinstance(var, type_):
        yield var
    for child in var.walk():
        for subvar in walk(child, type_):
            yield subvar


def hyperslab(slices):
    """
    Build an Opendap representation of a multidimensional slice object.

        >>> print hyperslab( (slice(None),) )
        <BLANKLINE>
        >>> print hyperslab( (slice(0, 10, 2),) )
        [0:2:9]

    Note that in Opendap the stop is included, while in Python it's
    exluded, as we can see on the last example. More examples::

        >>> print hyperslab( (slice(0, 10, 2), slice(1, 20, 2)) )
        [0:2:9][1:2:19]
        >>> print hyperslab( (slice(0, 10, 2), slice(None), slice(1, 20, 2)) )
        [0:2:9][0:1:2147483646][1:2:19]
        >>> print hyperslab( (slice(0, 10, 2), slice(1, 20, 2), slice(None)) )
        [0:2:9][1:2:19]

    For 1D slices we can omit the tuple::

        >>> print hyperslab(slice(0, 10, 2))
        [0:2:9]

    """
    if not isinstance(slices, tuple): slices = [slices]
    else: slices = list(slices)

    # Remove unnecessary slices from the end.
    while slices and slices[-1] == slice(None):
        slices.pop()

    return ''.join('[%d:%d:%d]' % (
            slice_.start or 0,
            slice_.step or 1,
            (slice_.stop or sys.maxint)-1)
            for slice_ in slices)


def fix_slice(slices, shape):
    """
    Fix a slice, handling ``None``s and integers.

    """
    if not isinstance(slices, tuple): slices = (slices,)
    n = len(shape)-len(slices)
    if [s for s in slices if s is Ellipsis]:  # ``Ellipsis in slices`` fails due to numpy comparison problem
        i = list(slices).index(Ellipsis)
        slices = slices[:i] + (slice(None),)*(n+1) + slices[i+1:]
    else:
        slices = slices + (slice(None),)*n

    out = []
    for slice_, length in zip(slices, shape):
        if isinstance(slice_, numpy.ndarray):
            start = slice_.nonzero()[0][0]
            stop = slice_.nonzero()[0][-1] + 1
            slice_ = slice(start, stop, 1)
        elif isinstance(slice_, (int, long)):
            if slice_ < 0: slice_ += length
            slice_ = slice(slice_, slice_+1, 1)
        else:
            start = slice_.start or 0
            if start < 0: start += length
            step = slice_.step or 1
            stop = slice_.stop or length
            if stop < 0: stop += length
            slice_ = slice(start, stop, step)
        out.append(slice_)
    return tuple(out)


def combine_slices(slices1, slices2):
    """
    Combine two multidimensional slices together.

    """
    out = []
    for slice1, slice2 in zip(slices1, slices2):
        if slice1.start is None and slice2.start is None:
            start = None
        else:
            start = (slice1.start or 0) + (slice2.start or 0)
        if slice1.step is None and slice2.step is None:
            step = None
        else:
            step = (slice1.step or 1) * (slice2.step or 1)
        if slice1.stop is None and slice2.stop is None:
            stop = None
        else:
            stop1 = slice1.stop or sys.maxint
            stop2 = slice2.stop or sys.maxint
            stop = min(stop1, (slice1.start or 0) + stop2)
        out.append(slice(start, stop, step))
    return tuple(out)


def get_slice(hyperslab):
    """
    Parse a hyperslab into a Python tuple of slices.

    """
    output = []
    dimslices = [ds for ds in hyperslab[1:-1].split('][') if ds]
    for dimslice in dimslices:
        tokens = dimslice.split(':')
        start = int(tokens[0])
        step = 1
        stop = start
        if len(tokens) == 2:
            stop = int(tokens[1])
        elif len(tokens) == 3:
            step = int(tokens[1])
            stop = int(tokens[2])
        output.append(slice(start, stop+1, step))
    return tuple(output)


def parse_qs(query):
    """
    Parse the constraint expression.

    """
    selection = []
    projection = []

    if query:
        selection = [urllib.unquote(token) for token in query.split('&')]
        if not re.search('<=|>=|!=|=~|>|<|=', selection[0]): 
            projection = [p for p in selection.pop(0).split(',') if p]

    fields = []
    for var in projection:
        tokens = var.split('.')
        tokens = [re.match('(.*?)(\[.*\])?$', token).groups() for token in tokens]
        tokens = [(token, get_slice(slice_ or '')) for (token, slice_) in tokens]
        fields.append(tokens)

    return fields, selection


def fix_shn(projection, dataset):
    """
    Fix shorthand notation for variables.
    
    Shorthand notation is the syntax some clients use to retrieve data
    using the variable name instead of its fully qualified id.

    """
    out = []
    for var in projection:
        if len(var) == 1 and var[0][0] not in dataset:
            token, slice_ = var.pop(0)
            for child in walk(dataset):
                if token == child.name:
                    if var: raise ConstraintExpressionError(
                            "Ambiguous shorthand notation request: %s" % token)
                    var = [(parent, ()) 
                            for parent in child.id.split('.')[:-1]] + [(token, slice_)]
        out.append(var)
    return out


def escape_dods(dods, pad=''):
    """
    Escape a DODS response.

    This is useful for debugging. You're probably spending too much time
    with pydap if you need to use this.

    """
    dds, dods = dods.split('\nData:\n', 1)

    out = []
    for i, char in enumerate(dods):
        char = hex(ord(char))
        char = char.replace('0x', '\\x')
        if len(char) < 4: char = char.replace('\\x', '\\x0')
        out.append(char)
        if pad and (i%4 == 3): out.append(pad)
    out = ''.join(out)
    out = out.replace(r'\x5a\x00\x00\x00', '<start of sequence>')
    out = out.replace(r'\xa5\x00\x00\x00', '<end of sequence>')
    return dds + '\nData:\n' + out



"""
DAP exceptions.

These exceptions are mostly used by the server. When an exception is
captured, a proper error message is displayed (according to the DAP
2.0 spec), with information about the exception and the error code 
associated with it.

The error codes are attributed using the "first come, first serve"
algorithm.

"""


class DapError(Exception):
    """
    Base DAP exception.
    
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class ClientError(DapError):
    """
    Generic error with the client.
    
    """
    code = 100


class ServerError(DapError):
    """
    Generic error with the server.
    
    """
    code = 200


class ConstraintExpressionError(ServerError):
    """
    Exception raised when an invalid constraint expression is given.
    
    """
    code = 201


class HandlerError(DapError):
    """
    Generic error with a handler.
    
    """
    code = 300


class ExtensionNotSupportedError(HandlerError):
    """
    Exception raised when trying to open a file not supported by any
    handlers.
    
    """
    code = 301


class OpenFileError(HandlerError):
    """
    Exception raised when unable to open a file.
    
    """
    code = 302




import itertools
from urllib import quote

import numpy

from pydap.model import *
from pydap.lib import INDENT, encode_atom
from pydap.responses.dds import dispatch as dds_dispatch
from pydap.responses.lib import BaseResponse


class ASCIIResponse(BaseResponse):
    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend([
                ('Content-description', 'dods_ascii'),
                ('Content-type', 'text/plain'),
                ])

    @staticmethod
    def serialize(dataset):
        # Generate DDS.
        for line in dds_dispatch(dataset):
            yield line
        yield 45 * '-'
        yield '\n'
        for line in dispatch(dataset):
            yield line
        if hasattr(dataset, 'close'): dataset.close()


def dispatch(var, printname=True):
    dispatchers = [(SequenceType, _sequence),
                   (StructureType, _structure),
                   (BaseType, _base)]
    
    for klass, func in dispatchers:
        if isinstance(var, klass):
            return func(var, printname)


def _structure(var, printname):
    for child in var.walk():
        for line in dispatch(child, printname):
            yield line
        yield '\n'


def _sequence(var, printname):
    yield ', '.join(child.id for child in var.values())
    yield '\n'
    for struct_ in var:
        out = []
        for child in struct_:
            for line in dispatch(child, printname=False):
                out.append(line)
        out = ', '.join(out).split('\n')
        for line in out:
            if line:
                yield line.strip(', ') + '\n'


def _base(var, printname):
    if printname:
        yield quote(var.id)
        yield '\n'

    data = numpy.asarray(var.data).flat
    if var.shape:
        for indexes, value in itertools.izip(numpy.ndindex(var.shape), data):
            index = '[%s]' % ']['.join(str(idx) for idx in indexes)
            yield '%s %s\n' % (index, encode_atom(value))
    else:
        yield encode_atom(var.data)



from pydap.lib import __version__, __dap__
from pydap.responses.lib import BaseResponse


class VersionResponse(BaseResponse):
    def __init__(self, dataset=None):
        BaseResponse.__init__(self, dataset)
        self.headers.extend([
                ('Content-description', 'dods_version'),
                ('Content-type', 'text/plain'),
                ])

    @staticmethod
    def serialize(dataset):
        output = """Core version: dods/%s
Server version: pydap/%s
""" % ('.'.join(str(d) for d in __dap__), '.'.join(str(d) for d in __version__))

        if hasattr(dataset, 'close'): dataset.close()
        return [output]




# This Python file uses the following encoding: utf-8
import struct

import numpy

from pydap.model import *


START_OF_SEQUENCE = '\x5a\x00\x00\x00'
END_OF_SEQUENCE = '\xa5\x00\x00\x00'


class DapPacker(object):
    def __init__(self, var):
        self.var = var

    def __iter__(self):
        if isinstance(self.var, SequenceType):
            for struct_ in self.var:
                yield START_OF_SEQUENCE
                for line in DapPacker(struct_):
                    yield line
            yield END_OF_SEQUENCE
        elif isinstance(self.var, StructureType):
            for child in self.var.walk():
                for line in DapPacker(child):
                    yield line
        else:
            # Make data iterable and put in blocks.
            if hasattr(self.var.data, 'shape'):
                data = self.var.data
            else:
                data = numpy.asarray(self.var.data)
            if len(data.shape) < 2:
                try: data.shape = (1,-1)
                except: pass  # arrayterator.shape is ro

            # Yield length (twice) if array.
            if getattr(self.var, 'shape', None):
                if self.var.type in [Url, String]:
                    yield self._pack_length()
                else:
                    yield self._pack_length() * 2

            # Bytes are sent differently.
            if self.var.type == Byte:
                for b in self._pack_bytes(data):
                    yield b
            # String are zero padded to 4n.
            elif self.var.type in [Url, String]:
                for block in data:
                    for word in block.flat:
                        yield self._pack_string(word)
            else:
                for block in data:
                    dtype = ">%s%s" % (self.var.type.typecode, self.var.type.size)
                    yield block.astype(dtype).tostring()

    def __str__(self):
        return ''.join(self)

    def _pack_length(self):
        shape = getattr(self.var, 'shape', [1])
        length = numpy.prod(shape)
        return struct.pack('>L', length)

    def _pack_bytes(self, data):
        count = 0
        for block in data:
            data = block.astype('B').tostring()
            yield data
            count -= len(data)
        yield (count % 4) * '\0'

    def _pack_string(self, s):
        """
        Pack a string.
        
        We first pack the string length, followed by the string padded
        to size 4n.

        """
        # Pack length first.
        n = len(s)
        length = struct.pack('>L', n)

        padding = -n % 4
        data = length + s + (padding * '\0')
        return data


class DapUnpacker(object):
    def __init__(self, xdrdata, var):
        self._buf = xdrdata
        self.var = var
        self._pos = 0

    def getvalue(self):
        if isinstance(self.var, SequenceType):
            out = []
            mark = self._unpack_uint()
            while mark == 1509949440:
                var = self.var
                # Create a structure with the sequence vars:
                self.var = StructureType(name=self.var.name)
                self.var.update(var)
                out.append(self.getvalue())
                self.var = var
                mark = self._unpack_uint()

        elif isinstance(self.var, StructureType):
            out = []
            for child in self.var.walk():
                var = self.var
                self.var = child
                out.append(self.getvalue())
                self.var = var
            out = tuple(out)

        else:
            # Get data length.
            n = 1
            if getattr(self.var, 'shape', False):
                n = self._unpack_uint()
                if self.var.type not in [Url, String]:
                    self._unpack_uint()
                
            # Bytes are treated differently.
            if self.var.type == Byte:
                out = self._unpack_bytes(n)
                out = numpy.array(out, self.var.type.typecode)
            # As are strings...
            elif self.var.type in [Url, String]:
                out = self._unpack_string(n)
                out = numpy.array(out, self.var.type.typecode)
            else:
                i = self._pos
                self._pos = j = i + (n*self.var.type.size)
                dtype = ">%s%s" % (self.var.type.typecode, self.var.type.size)
                out = numpy.fromstring(self._buf[i:j], dtype=dtype)

            if getattr(self.var, 'shape', False):
                out.shape = tuple(idx for idx in self.var.shape if idx != 1)
            else:   
                out = out[0]

        return out

    def _unpack_uint(self):
        i = self._pos
        self._pos = j = i+4
        data = self._buf[i:j]
        if len(data) < 4:
            raise EOFError
        x = struct.unpack('>L', data)[0]
        try:
            return int(x)
        except OverflowError:
            return x

    def _unpack_bytes(self, count):
        i = self._pos
        self._pos = j = i+count
        out = numpy.fromstring(self._buf[i:j], dtype='B')
        padding = -count % 4
        self._pos += padding
        return out

    def _unpack_string(self, count):
        out = []
        for s in range(count):
            # Unpack string length.
            n = self._unpack_uint()

            i = self._pos
            self._pos = j = i+n
            data = self._buf[i:j]
            out.append(data)

            # Fix cursor position.
            padding = -n % 4
            self._pos += padding
        return out



import copy
import re
import operator

from pydap.model import *
from pydap.lib import walk, parse_qs, fix_shn
from pydap.util.safeeval import expr_eval


def constrain(dataset, ce):
    """
    A constraint expression applier.

        >>> dataset = DatasetType(name='test')
        >>> dataset['seq'] = SequenceType(name='seq')
        >>> dataset['seq']['index'] = BaseType(name='index', type=Int32)
        >>> dataset['seq']['temperature'] = BaseType(name='temperature', type=Float32)
        >>> dataset['seq']['site'] = BaseType(name='site', type=String)

        >>> dataset['seq'].data = [
        ...         (10, 17.2, 'Diamont_St'),
        ...         (11, 15.1, 'Blacktail_Loop'),
        ...         (12, 15.3, 'Platinum_St'),
        ...         (13, 15.1, 'Kodiak_Trail')]
        >>> for struct_ in dataset.seq:
        ...     print struct_.data
        (10, 17.199999999999999, 'Diamont_St')
        (11, 15.1, 'Blacktail_Loop')
        (12, 15.300000000000001, 'Platinum_St')
        (13, 15.1, 'Kodiak_Trail')

        >>> dataset2 = constrain(dataset, 'seq.index>11')
        >>> for struct_ in dataset2.seq:
        ...     print struct_.data
        (12, 15.300000000000001, 'Platinum_St')
        (13, 15.1, 'Kodiak_Trail')
        >>> dataset2 = constrain(dataset, 'seq.index>11&seq.temperature<15.2')
        >>> for struct_ in dataset2.seq:
        ...     print struct_.data
        (13, 15.1, 'Kodiak_Trail')

        >>> dataset.clear()
        >>> dataset['casts'] = SequenceType(name='casts')
        >>> dataset['casts']['lat'] = BaseType(name='lat', type=Float32)
        >>> dataset['casts']['lon'] = BaseType(name='lon', type=Float32)
        >>> dataset['casts']['time'] = BaseType(name='time', type=Float64)
        >>> dataset['casts']['profile'] = SequenceType(name='profile')
        >>> dataset['casts']['profile']['t'] = BaseType(name='t', type=Float32)
        >>> dataset['casts']['profile']['s'] = BaseType(name='s', type=Float32)
        >>> dataset['casts']['profile']['p'] = BaseType(name='p', type=Float32)
        >>> dataset['casts'].data = [
        ...         (-10.0, 290.0, 1.0, [(21.0, 35.0, 100.0), (20.5, 34.9, 200.0), (19.0, 33.0, 300.0)]),
        ...         (-11.0, 295.0, 2.0, [(22.0, 35.5, 100.0), (21.0, 35.4, 200.0), (20.0, 33.5, 300.0), (19.0, 33.0, 500.0)])]
        >>> dataset2 = constrain(dataset, 'casts.lat>-11')
        >>> for struct_ in dataset2.casts:
        ...     print struct_.data
        (-10.0, 290.0, 1.0, array([[21.0, 35.0, 100.0],
               [20.5, 34.9, 200.0],
               [19.0, 33.0, 300.0]], dtype=object))

    Filtering is guaranteed to work only in outer sequences, but not in inner
    sequences like this::

        >>> dataset2 = constrain(dataset, 'casts.profile.p>100')
        >>> for struct_ in dataset2.casts:
        ...     print struct_.data
        (-10.0, 290.0, 1.0, array([[21.0, 35.0, 100.0],
               [20.5, 34.9, 200.0],
               [19.0, 33.0, 300.0]], dtype=object))
        (-11.0, 295.0, 2.0, array([[22.0, 35.5, 100.0],
               [21.0, 35.4, 200.0],
               [20.0, 33.5, 300.0],
               [19.0, 33.0, 500.0]], dtype=object))
        
    Instead, inner sequences have to be filtered inside a loop::

        >>> for struct_ in dataset.casts:
        ...     for profile in struct_.profile[ struct_.profile.p > 100 ]:
        ...         print profile.data
        (20.5, 34.899999999999999, 200.0)
        (19.0, 33.0, 300.0)
        (21.0, 35.399999999999999, 200.0)
        (20.0, 33.5, 300.0)
        (19.0, 33.0, 500.0)

    """
    projection, selection = parse_qs(ce)
    projection = projection or [[(key, ())]
            for key in dataset.keys()]
    projection = fix_shn(projection, dataset)

    # Make a copy of the dataset.
    filtered = copy.deepcopy(dataset)

    # Filter sequences.
    for seq in walk(filtered, SequenceType):
        if seq._nesting_level == 1:
            filter_ = []
            # Check only selections that apply to the direct children of this sequence
            # (ie, skip children from nested sequences).
            for cond in [cond for cond in selection
                    if re.match('%s\.[^\.]+(<=|<|>=|>|=|!=)' % re.escape(seq.id), cond)]:
                id_, op, other = parse_selection(cond, dataset)
                filter_.append(op(id_, other))
            if filter_:
                seq.data = seq[ reduce(lambda c1, c2: c1 & c2, filter_) ].data

    # Create a new empty dataset to build it up.
    new_ = DatasetType(name=filtered.name,
            attributes=filtered.attributes.copy())

    for var in projection:
        target, template = new_, filtered
        while var:
            name, slice_ = var.pop(0)
            candidate = copy.deepcopy(template[name])
            if slice_:
                if isinstance(candidate, SequenceType):
                    candidate = candidate[slice_[0]]
                elif isinstance(candidate, BaseType):
                    candidate.data = candidate[slice_]
                    candidate.shape = candidate.data.shape
                else:
                    candidate = candidate[slice_] 

            if isinstance(candidate, StructureType):
                if var:
                    # Convert degenerate grids into structures.
                    if isinstance(candidate, GridType):
                        candidate.__class__ = StructureType
                    candidate.clear()
                if name not in target or not var:
                    target[name] = candidate
                target, template = target[name], template[name]
            else:
                target[name] = candidate

    return new_


def parse_selection(cond, dataset):
    """
    Convert a URL selection to 3 tokens.

    """
    id_, op, other = re.split('(<=|>=|!=|=~|>|<|=)', cond)

    op = { '<=': operator.le, '>=': operator.ge,
            '!=': operator.ne, '=': operator.eq,
            '>': operator.gt, '<': operator.lt }[op]
    try:
        names = [dataset] + id_.split('.')
        id_ = reduce(operator.getitem, names)
    except:
        id_ = expr_eval(id_)
    try:
        names = [dataset] + other.split('.')
        other = reduce(operator.getitem, names)
    except:
        other = expr_eval(other)

    return id_, op, other


def _test():
    import doctest
    doctest.testmod()


if __name__ == "__main__":
    _test()




import sys
import re
from urllib import unquote
import logging

from pkg_resources import iter_entry_points
from paste.httpexceptions import HTTPException
from paste.deploy.converters import asbool

from pydap.lib import parse_qs
from pydap.handlers.helper import constrain
from pydap.responses.error import ErrorResponse
from pydap.exceptions import ExtensionNotSupportedError


def load_handlers():
    return [ep.load() for ep in iter_entry_points("pydap.handler")]


def get_handler(filepath, handlers=None):
    # Check each handler to see which one handles this file.
    for handler in handlers or load_handlers():
        p = re.compile(handler.extensions)
        if p.match(filepath):
            return handler(filepath)

    raise ExtensionNotSupportedError(
            'No handler available for file %s.' % filepath)


class BaseHandler(object):
    """
    Base class for pydap handlers.

    This class is the base class for the different handlers. The job
    of building a given dataset from a given request is passed to the
    ``parse_constraints`` method, which should be defined in each
    specific handler. The dataset is then passed to a response class,
    which is simply a WSGI application that is returned.

    """

    # Load all the different supported responses (DDS/DAS/etc.).
    response_map = dict((r.name, r.load())
            for r in iter_entry_points('pydap.response'))

    def __call__(self, environ, start_response):
        path_info = environ.get('PATH_INFO') or environ.get('SCRIPT_NAME', '')
        path_info = unquote(path_info).lstrip('/')
        path, type_ = path_info.rsplit('.', 1)
        # For DAS, we need to clear the constraint expression to get the
        # full dataset.
        if type_ == 'das': environ['QUERY_STRING'] = ''

        # Add some vars to the environment.
        environ['pydap.path'] = path
        environ['pydap.response'] = type_
        environ['pydap.ce'] = parse_qs(environ.get('QUERY_STRING', ''))
        environ['pydap.logger'] = logging.getLogger('pydap')
        
        try:
            # Build the dataset using the proper subclass method and
            # pass it to a response (DAS/DDS/etc.) builder.
            dataset = self.parse_constraints(environ)
            response = self.response_map[type_]
            responder = response(dataset)
            return responder(environ, start_response)
        except HTTPException, exc:
            # Some responses (like HTML) raise an HTTPRedirect when the
            # form is posted, so we need this here.
            return exc(environ, start_response)
        except:
            # We silently capture and format exceptions, unless this
            # environ flag says otherwise.
            if asbool(environ.get('x-wsgiorg.throw_errors')):
                raise
            return ErrorResponse(info=sys.exc_info())(environ, start_response)
        
    def parse_constraints(self, environ):
        raise NotImplementedError(
                'Subclasses must implement parse_constraints')


class SimpleHandler(BaseHandler):
    """
    A simple handler for ``dap.model.DatasetType`` datasets.

    This handler takes care of datasets built by hand from the types in
    ``pydap.model``.

    """
    def __init__(self, dataset, debug=False):
        self.dataset = dataset
        self.debug = debug

    def parse_constraints(self, environ):
        """
        Parse the request.

        This method uses the magic ``constrain`` function, which handles
        arbitrary datasets/constraint expressions according to the DAP spec.

        """
        environ['x-wsgiorg.throw_errors'] = self.debug
        new_dataset = constrain(self.dataset, environ.get('QUERY_STRING', ''))
        if hasattr(self.dataset, 'close'): self.dataset.close()
        return new_dataset



from paste.request import construct_url

from pydap.lib import __version__
from pydap.util.template import GenshiRenderer, StringLoader
from pydap.responses.lib import BaseResponse


DEFAULT_TEMPLATE = """<html xmlns="http://www.w3.org/1999/xhtml">
    <head>
        <title>pyDAP help page for file ${location[:-5]}</title>

        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    </head>

    <body>
        <div id="main">
            <h1>How to access this file</h1>

            <p>The resource <code>${location[:-5]}</code> is a dataset available through the <a href="http://opendap.org/">OPeNDAP</a> protocol. The best way to access this dataset is using an <a href="http://www.opendap.org/faq/whatClients.html">OPeNDAP client</a> like Ferret, GrADS or Matlab.</p>

            <p>In case you <strong>don't have</strong> a client installed, you can also download data by accessing the <a href="${location[:-5]}.html">web interface</a> for this dataset.</p>

            <hr />
            <p class="footnote"><em><a href="http://pydap.org/">pydap/$version</a></em></p>
        </div>
    </body>
</html>"""


class HelpResponse(BaseResponse):

    renderer = GenshiRenderer(
            options={}, loader=StringLoader( {'help.html': DEFAULT_TEMPLATE} ))

    def __init__(self, dataset=None):
        BaseResponse.__init__(self, dataset)
        self.headers.extend([
                ('Content-description', 'dods_help'),
                ('Content-type', 'text/html'),
                ])

    def __call__(self, environ, start_response):
        def serialize(dataset):
            location = construct_url(environ, with_query_string=False)
            base = location[:location.rfind('/')]
            context = {
                'root': construct_url(environ, with_query_string=False, script_name='').rstrip('/'),
                'base': base,
                'location': location,
                'version' : '.'.join(str(d) for d in __version__),
            }
            renderer = environ.get('pydap.renderer', self.renderer)
            template = renderer.loader('help.html')
            output = renderer.render(template, context)
            if hasattr(dataset, 'close'): dataset.close()
            return [output]
        self.serialize = serialize
        return BaseResponse.__call__(self, environ, start_response)



"""
A simple file-based Opendap server.

Serves files from a root directory, handling those recognized by installed
handlers.

"""

import re
import os
from urllib import unquote
import time

from paste.urlparser import StaticURLParser
from paste.request import construct_url, path_info_pop

from pydap.handlers.lib import get_handler, load_handlers
from pydap.lib import __version__
from pydap.exceptions import ExtensionNotSupportedError
from pydap.util.template import FileLoader, GenshiRenderer


class FileServer(StaticURLParser):
    def __init__(self, root, root_directory=None, cache_max_age=None, templates='templates', catalog='catalog.xml', **config):
        StaticURLParser.__init__(self, root, root_directory, cache_max_age)
        self.catalog = catalog
        self.config = config

        self.templates = templates
        loader = FileLoader(templates)
        self.renderer = GenshiRenderer(
                options={}, loader=loader)

        self.handlers = load_handlers()

    def __call__(self, environ, start_response):
        path_info = environ.get('PATH_INFO', '')
        if not path_info:
            return self.add_slash(environ, start_response)
        if path_info == '/':
            # @@: This should obviously be configurable
            filename = 'index.html'
        else:
            filename = path_info_pop(environ)
        full = os.path.normcase(os.path.normpath(
            os.path.join(self.directory, filename)))
        if os.path.sep != '/':
            full = full.replace('/', os.path.sep)
        if self.root_directory is not None and not full.startswith(self.root_directory):
            # Out of bounds
            return self.not_found(environ, start_response)
        if not os.path.exists(full):
            return self.not_found(environ, start_response)
        if os.path.isdir(full):
            # @@: Cache?
            child_root = self.root_directory is not None and \
                self.root_directory or self.directory
            return self.__class__(full, root_directory=child_root,
                                  cache_max_age=self.cache_max_age,
                                  templates=self.templates,
                                  catalog=self.catalog,
                                  **self.config)(environ, start_response)
        if environ.get('PATH_INFO') and environ.get('PATH_INFO') != '/':
            return self.error_extra_path(environ, start_response)
        if_none_match = environ.get('HTTP_IF_NONE_MATCH')
        if if_none_match:
            mytime = os.stat(full).st_mtime
            if str(mytime) == if_none_match:
                headers = []
                ETAG.update(headers, mytime)
                start_response('304 Not Modified', headers)
                return [''] # empty body

        fa = self.make_app(full)
        if self.cache_max_age:
            fa.cache_control(max_age=self.cache_max_age)
        return fa(environ, start_response)

    def not_found(self, environ, start_response, debug_message=None):
        request = os.path.split(environ['SCRIPT_NAME'])[1]
        request = os.path.normpath(os.path.join(self.directory, request))
        filepath, response = os.path.splitext(request)
        if os.path.sep != '/': filepath = filepath.replace('/', os.path.sep)

        if environ['PATH_INFO'] == '/':
            return self.index(environ, start_response,
                    'index.html', 'text/html')
        elif os.path.exists(filepath):
            # Update environ with configuration keys (environ wins in case of conflict).
            for k in self.config:
                environ.setdefault(k, self.config[k])
            environ['pydap.renderer'] = self.renderer
            handler = get_handler(filepath, self.handlers)
            return handler(environ, start_response)
        elif environ['SCRIPT_NAME'].endswith('/%s' % self.catalog):
            return self.index(environ, start_response,
                    'catalog.xml', 'text/xml')
        else:
            return StaticURLParser.not_found(
                    self, environ, start_response, debug_message)

    def index(self, environ, start_response, template_name, content_type):
        # Return directory listing.
        dirs = []
        files = []
        for name in os.listdir(self.directory):
            path = os.path.normpath(os.path.join(self.directory, name))
            if os.path.isdir(path) and not name.startswith('.'):
                dirs.append(name)
            elif os.path.isfile(path) and not name.startswith('.'):
                statinfo = os.stat(path)
                file = {'name': name,
                        'size': format_size(statinfo[6]),
                        'modified': time.localtime(statinfo[8]),
                        'supported': False}
                try:
                    get_handler(path, self.handlers)
                    file['supported'] = True
                except ExtensionNotSupportedError:
                    pass
                files.append(file)
        # Sort naturally using Ned Batchelder's algorithm.
        dirs.sort(key=alphanum_key)
        files.sort(key=lambda l: alphanum_key(l['name']))

        # Base URL.
        location = construct_url(environ, with_query_string=False)
        base = location[:location.rfind('/')]

        context = {
                'root': construct_url(environ, with_query_string=False, script_name='').rstrip('/'),
                'base': base,
                'location': location,
                'title': 'Index of %s' % unquote(environ.get('SCRIPT_NAME') or '/'),
                'dirs' : dirs,
                'files': files,
                'version': '.'.join(str(d) for d in __version__)
        }
        template = self.renderer.loader(template_name)
        output = self.renderer.render(template, context)
        headers = [('Content-type', content_type)]
        start_response("200 OK", headers)
        return [output]


# http://svn.colorstudy.com/home/ianb/ImageIndex/indexer.py
def format_size(size):
    if not size:
        return 'empty'
    if size > 1024:
        size = size / 1024.
        if size > 1024:
            size = size / 1024.
            return '%.1i MB' % size
        return '%.1f KB' % size
    return '%i bytes' % size


def alphanum_key(s):
    """
    Turn a string into a list of string and number chunks.

        >>> alphanum_key("z23a")
        ['z', 23, 'a']

    From http://nedbatchelder.com/blog/200712.html#e20071211T054956

    """
    def tryint(s):
        try:
            return int(s)
        except:
            return s
    return [tryint(c) for c in re.split('([0-9]+)', s)]


def make_app(global_conf, root, templates, **kwargs):
    return FileServer(root, templates=templates, **kwargs)



from urllib import quote

from pydap.model import *
from pydap.lib import INDENT
from pydap.responses.lib import BaseResponse


class DDSResponse(BaseResponse):
    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend([
                ('Content-description', 'dods_dds'),
                ('Content-type', 'text/plain'),
                ])

    @staticmethod
    def serialize(dataset):
        output = ''.join(dispatch(dataset))
        if hasattr(dataset, 'close'): dataset.close()
        return [output]


def dispatch(var, level=0):
    dispatchers = [(DatasetType, structure_builder('Dataset')),
                   (SequenceType, structure_builder('Sequence')),
                   (GridType, _grid),
                   (StructureType, structure_builder('Structure')),
                   (BaseType, _base)]

    for klass, func in dispatchers:
        if isinstance(var, klass):
            return func(var, level)


def structure_builder(name):
    def func(var, level=0):
        yield '%s%s {\n' % (level * INDENT, name)

        # Get the DDS from stored variables.
        for child in var.walk():
            for line in dispatch(child, level=level+1):
                yield line
        yield '%s} %s;\n' % (level * INDENT, quote(var.name))
    return func


def _grid(var, level=0):
    yield '%sGrid {\n' % (level * INDENT)

    yield '%sArray:\n' % ((level+1) * INDENT)
    for line in _base(var.array, level=level+2):
        yield line

    yield '%sMaps:\n' % ((level+1) * INDENT)
    for map_ in var.maps.values():
        for line in _base(map_, level=level+2):
            yield line

    yield '%s} %s;\n' % (level * INDENT, quote(var.name))


def _base(var, level=0):
    if var.dimensions:
        dims = ['%s = %d' % dim for dim in zip(var.dimensions, var.shape)]
    else:
        if len(var.shape) == 1:
            dims = ['%s = %d' % (quote(var.name), var.shape[0])]
        else:
            dims = ['%d' % i for i in var.shape]
    shape = dims and '[%s]' % ']['.join(dims) or ''
    yield '%s%s %s%s;\n' % (level * INDENT, var.type.descriptor, quote(var.name), shape)



from operator import attrgetter, itemgetter
import copy
import itertools
from new import classobj

import numpy

from pydap.lib import quote, walk, fix_slice
from pydap.util.odict import odict


__all__ = ['StructureType', 'SequenceType', 'DatasetType', 'GridType',
           'BaseType', 'Float32', 'Float64', 'Int16', 'Int32', 'UInt16',
           'UInt32', 'Byte', 'String', 'Url', 'typemap']


# Define the basic Opendap types as classes. Each class has a correspondent
# Numpy typecode and item size. Instead of explicitly defining classes we
# can use the ``classobj`` function, since the classes are basically just
# placeholders for attributes.
def TypeFactory(name, typecode, size):
    return classobj(name, (object,),
            {'typecode': typecode, 'size': size, 'descriptor': name})

Float64 = TypeFactory('Float64', 'd', 8)
Float32 = TypeFactory('Float32', 'f', 4)
Int32 = TypeFactory('Int32', 'i', 4)
Int16 = TypeFactory('Int16', 'i', 4)
UInt32 = TypeFactory('UInt32', 'I', 4)
UInt16 = TypeFactory('UInt16', 'I', 4)
Byte = TypeFactory('Byte', 'B', 1)
String = TypeFactory('String', 'S', None)
Url = TypeFactory('Url', 'S', None)
basetypes = [Float64, Float32, Int32, Int16, Byte, UInt32, UInt16, String, Url]

# A simple map to convert between commonly used identifiers
# and our classes.
typemap = {
    # type.__name__.lower()
    'float64': Float64,
    'float32': Float32,
    'int32': Int32,
    'int16': Int16,
    'uint32': UInt32,
    'uint16': UInt16,
    'byte': Byte,
    'string': String,
    'url': Url,

    # numpy
    'd': Float64,
    'f': Float32,
    'h': Int16,
    'i': Int32, 'l': Int32, 'q': Int32,
    'b': Byte,
    'H': UInt16,
    'I': UInt32, 'L': UInt32, 'Q': UInt32,
    'B': Byte,
    'S': String,

    # extra from the stdlib array module
    'c': String,
    'u': String,
}


class DapType(object):
    """
    The common Opendap type.

    This class is an abstract class, defining common methods and
    attributes for other all classes in the Opendap data model.

    """
    def __init__(self, name='nameless', attributes=None, nesting_level=0):
        self.name = name
        self._id = name
        self.attributes = attributes or {}
        self._nesting_level = nesting_level

    # DAP variables' names cannot have special characters like
    # [, ] or %. Here we automatically escape them when the name
    # is set.
    def _get_name(self):
        return self._name
    def _set_name(self, name):
        self._name = quote(name)
    name = property(_get_name, _set_name)

    def __getattr__(self, attr):
        """
        Attribute shortcut.

        The data classes have their attributes stored in the 
        ``attributes`` attribute, which is a dictionary. Access
        to these values can be shortcutted by accessing the 
        attribute directly::

            >>> var = DapType()
            >>> var.attributes['foo'] = 'bar'
            >>> print var.foo
            bar

        """
        try:
            return self.attributes[attr]
        except (KeyError, TypeError):
            raise AttributeError(
                    "'%s' object has no attribute '%s'"
                    % (self.__class__, attr))
    
    def walk(self):
        """
        Iterate over children.

        This method is used in constructor variables to iterate
        over the variable's children. The default behavior is to
        return and empty iterable (ie, no children).

        """
        return ()

    def _set_id(self, parent=None):
        """
        Set the variable id.

        The id of a variable is a representation of its hierarchy
        in the dataset, using dots to join the variable names.

        """
        if parent:
            self._id = '%s.%s' % (parent, self.name)
        else:
            self._id = self.name

        # Propagate id to children.
        for var in self.walk():
            var._set_id(self._id)
    id = property(attrgetter('_id'))  # read-only


class BaseType(DapType):
    """
    The base Opendap type.

    This class represents basic Opendap types, which contain data.
    Variables can be scalars or multi-dimensional arrays, of one of
    the basic Opendap types (Int32, String, etc.)::

        >>> a = BaseType(name='a', data=1, type=Int32)
        >>> b = BaseType(name='b', data=range(5), shape=(5,),
        ...         dimensions=('time',), type=UInt16)
        >>> print b.attributes
        {}
        >>> b.attributes['units'] = 'days since 1980-1-1'

    Comparisons and other operators are usually applied to the ``data``
    attribute::

        >>> print b[1]
        1
        >>> print len(b)
        5
        >>> print a > 0
        True
        >>> data = numpy.arange(4)
        >>> data.shape = (2, 2)
        >>> c = BaseType(name='c', data=data)
        >>> for block in c:
        ...     print block
        [0 1]
        [2 3]

    """
    def __init__(self, name='nameless', data=None, shape=None,
            dimensions=None, type=Int32, attributes=None):
        DapType.__init__(self, name, attributes)
        self.type = type in basetypes and type or typemap[type]
        self.data = data
        self.shape = shape or ()
        self.dimensions = dimensions or ()

    def __str__(self):
        return """%s
    with data
%s""" % (self.__class__, self.data)

    def __getitem__(self, index):
        return self.data[index]

    @property
    def __array_interface__(self):
        data = numpy.asarray(self.data)
        return {
                'version': 3,
                'shape': data.shape,
                'typestr': data.dtype.str,
                'data': data,
                }

    # Comparisons and other operations are applied directly to
    # the ``data`` attribute.
    def __eq__(self, other): return self.data == other
    def __ne__(self, other): return self.data != other
    def __ge__(self, other): return self.data >= other
    def __le__(self, other): return self.data <= other
    def __gt__(self, other): return self.data > other
    def __lt__(self, other): return self.data < other
    def __iter__(self): return iter(self.data)
    def __len__(self): return len(self.data)

    def __copy__(self):
        out = self.__class__(self.name, self.data, self.shape,
                self.dimensions, self.type, self.attributes)
        out._id = self._id
        return out

    def __deepcopy__(self, memo=None, _nil=[]):
        """
        Return a copy of the object, with a copy of the data too.

        """
        out = self.__class__(self.name, copy.copy(self.data),
                self.shape[:], self.dimensions[:], self.type,
                self.attributes.copy())
        out._id = self._id
        return out


class StructureType(odict, DapType):
    """
    An Opendap Structure.

    A StructureType is simply a fancy dictionary, used to hold groups of
    variables that share common characteristics (in theory).  They work
    exactly like Python dictionaries::

        >>> s = StructureType(name='s')
        >>> s['a'] = BaseType(name='a')
        >>> s['b'] = BaseType(name='b')
        >>> print s.keys()
        ['a', 'b']
        >>> print [var.name for var in s]
        ['a', 'b']

    """
    def __init__(self, name='nameless', attributes=None):
        odict.__init__(self)
        DapType.__init__(self, name, attributes)

    def __getattr__(self, attr):
        """
        Allow lazy access to children.

        We override ``__getattr__`` to allow children variables
        to be accessed using a lazy syntax::

            >>> s = StructureType(name='s')
            >>> s['a'] = BaseType(name='a')
            >>> print s.a.id
            s.a

        """
        if attr in self:
            return self[attr]
        else:
            return DapType.__getattr__(self, attr)

    # Walk returns each stored variable.
    walk = __iter__ = odict.itervalues

    def __setitem__(self, key, item):
        if key != item.name:
            raise KeyError('Key "%s" is different from variable name "%s"!' %
                    (key, item.name))
        odict.__setitem__(self, key, item)

        # Fix id in item.
        item._set_id(self._id)

    # Propagate to and collect data from children.
    def _get_data(self):
        data = [ var.data for var in self.walk() ]
        return tuple(
                combine_rows(data, self._nesting_level))
    def _set_data(self, data):
        data = [get_row(data, i, self._nesting_level) for
                i, k in enumerate(self.keys())]
        for col, var in itertools.izip(data, self):
            var.data = col
    data = property(_get_data, _set_data)

    def __copy__(self):
        out = self.__class__(self.name, self.attributes)
        out._id = self._id

        # Stored variables are not copied.
        out.update(self)
        return out

    def __deepcopy__(self, memo=None, _nil=[]):
        out = self.__class__(self.name, self.attributes.copy())
        out._id = self._id

        # Make copies of the stored variables.
        for k, v in self.items():
            out[k] = copy.deepcopy(v, memo)
        return out


class DatasetType(StructureType):
    """
    An Opendap Dataset.

    A DatasetType works pretty much like a Structure; the major
    difference is that its name is not used when composing the id of
    stored variables::

        >>> dataset = DatasetType(name='dataset')
        >>> dataset['a'] = BaseType(name='a', attributes={'foo': 'bar'})
        >>> print dataset.a.foo
        bar

    """
    def __setitem__(self, key, item):
        if key != item.name:
            raise KeyError('Key "%s" is different from variable name "%s"!' %
                    (key, item.name))
        odict.__setitem__(self, key, item)

        # Do not propagate id.
        item._set_id(None)

    def _set_id(self, parent=None):
        self._id = self.name

        for var in self.walk():
            var._set_id(None)


class SequenceType(StructureType):
    """
    An Opendap Sequence.

    Sequences are a special kind of constructor, holding records for
    the stored variables. They are somewhat similar to record arrays
    in Numpy::

        >>> s = SequenceType(name='s')
        >>> s['id'] = BaseType(name='id', type=Int32)
        >>> s['x'] = BaseType(name='x', type=Float64)
        >>> s['y'] = BaseType(name='y', type=Float64)
        >>> s['foo'] = BaseType(name='foo', type=Int32)

        >>> s.data = [(1, 10, 100, 42), (2, 20, 200, 43), (3, 30, 300, 44)]
        >>> for struct_ in s: print struct_.data
        (1, 10, 100, 42)
        (2, 20, 200, 43)
        (3, 30, 300, 44)
        >>> del s['foo']
        >>> print s.data
        [[1 10 100]
         [2 20 200]
         [3 30 300]]
        >>> print s['id'].data
        [1 2 3]

    Note that we had to use ``s['id']`` to refer to the variable ``id``,
    since ``s.id`` already points to the id of the Sequence.
    
    (An important point is that the ``data`` attribute must be copiable,
    so don't use consumable iterables like older versions of Pydap
    allowed.)

    Sequences are quite versatile; they can be indexed::

        >>> print s[0].data
        [[1 10 100]]
        >>> print s[0].x.data
        [10]

    Or filtered::

        >>> print s[ (s['id'] > 1) & (s.x > 10) ].data
        [[2 20 200]
         [3 30 300]]

    Or even both::

        >>> print s[ s['id'] > 1 ][1].x.data
        [30]

    If you mix indexing and filtering, be sure to use the right Sequence
    on the filter::

        >>> print s[ s['id'] > 1 ][1].x.data
        [30]
        >>> print s[1][ s['id'] > 1 ].x.data
        Traceback (most recent call last):
            ...
        IndexError: index (1) out of range (0<=index<0) in dimension 0
        >>> print s[1][ s[1]['id'] > 1 ].x.data
        [20]

    (Note that there's a difference between filtering first and then
    slicing, and slicing first and then indexing. This might not be the
    case always, since an Opendap server will always apply the filter
    first, while in this case we're working locally with the data. Don't
    worry, though: when this happens while accessing an Opendap server
    a warning will be issued by the client.)
    
    When filtering a Sequence, don't use the Python extended comparison
    syntax of ``1 < a < 2``, otherwise bad things will happen.

    And of course, slices are also used to access children::

        >>> print s['x'] is s.x
        True

    """
    def __init__(self, name='nameless', attributes=None, data=None):
        StructureType.__init__(self, name, attributes)
        self._nesting_level = 1
        self._data = None
        self.data = data

    def __setitem__(self, key, item):
        StructureType.__setitem__(self, key, item)

        # Increment ``_nesting_level`` in nested sequences.
        def increase_level(var):
            for child in var.walk():
                child._nesting_level = var._nesting_level
                if isinstance(child, SequenceType):
                    child._nesting_level += 1
                increase_level(child)
        increase_level(self)

        if self._data is not None:
            self.data = self.data[ tuple([var.name for var in self.walk()]) ]

    def __delitem__(self, key):
        StructureType.__delitem__(self, key)

        if self._data is not None:
            self.data = self.data[ tuple([var.name for var in self.walk()]) ]

    # When we set the Sequence data we also set the data for its children.
    def _set_data(self, data):
        if isinstance(data, SequenceData):
            self._data = data
            for var in walk(self, (BaseType, SequenceType)):
                if var is not self:
                    id_ = var.id[len(self.id)+1:]
                    var.data = data[id_]
        else:
            self._data = None
            for i, var in enumerate(self.walk()):
                var.data = numpy.asarray(
                        get_row(data, i, self._nesting_level), 'O')
    def _get_data(self):
        if self._data is not None:
            return self._data
        else:
            # Combine the data using the children's data.
            data = [ var.data for var in self.walk() ]
            return numpy.asarray(
                    combine_rows(data, self._nesting_level), 'O')
    data = property(_get_data, _set_data)

    def __getitem__(self, key):
        """
        Fancy Sequence slicing.

        The basic rule for Sequence slices is that it always return a new
        variable -- either a child or a new Sequence. To select a child,
        just use the common dictionary syntax::

            >>> s = SequenceType(name='s')
            >>> s['a'] = BaseType(name='a')
            >>> print s['a'].id
            s.a

        A Sequence can also be filtered or indexed using the slice::

            >>> s['b'] = BaseType(name='b')
            >>> s['c'] = BaseType(name='c')
            >>> s.data = [(1, 10, 100), (2, 20, 200), (3, 30, 300)]
            >>> print s[0].data
            [[1 10 100]]
            >>> print s[2:10].data
            [[3 30 300]]
            >>> print s[ s.a > 10 ].data
            []

        These will return new Sequence objects with the appropriate data.

        """
        if isinstance(key, basestring) and key in self:
            return odict.__getitem__(self, key)
        else:
            out = copy.deepcopy(self)
            if isinstance(key, tuple):
                # return a new sequence with selected children
                out._keys = list(key)
                if out._data is not None: out.data = out.data[key]  # SequenceData protocol
            else:
                if isinstance(key, (int, long)): key = slice(key, key+1, 1)
                out.data = out.data[key]
            return out

    def __iter__(self):
        """
        Return a Structure for each record in the Sequence.

        When a Sequence is iterated it returns Structure objects
        containing each record.

        """
        data = copy.deepcopy(self.data)
        data = list(data)  # download data before extracting columns
        cols = (get_row(data, i, self._nesting_level) for
                i, k in enumerate(self.keys()))
        for row in zip(*cols):
            struct_ = StructureType(self.name, self.attributes.copy())
            for col, name in zip(row, self.keys()):
                var = struct_[name] = copy.deepcopy(self[name])
                if isinstance(var, SequenceType):
                    var._nesting_level -= 1
                var.data = col
            yield struct_

    def __deepcopy__(self, memo=None, _nil=[]):
        out = StructureType.__deepcopy__(self, memo, _nil)
        out._nesting_level = self._nesting_level
        out._data = copy.copy(self._data)
        return out


class SequenceData(object):
    pass


def get_row(data, col, level):
    if level == 0:
        return data[col]
    else:
        return [ get_row(value, col, level-1) for value in data ]


def combine_rows(data, level):
    if level == 0:
        return data
    else:
        return [ combine_rows(value, level-1) for value in zip(*data) ]


class GridType(StructureType):
    """
    An Opendap Grid.

    A Grid works both like an array and a Structure. The Grid is basically
    a Structure containing an array and more variables describing its
    axes; the first defined variable is the multi-dimensional array,
    while later each individual axis should be defined::

        >>> g = GridType(name='g')
        >>> data = numpy.arange(6.)
        >>> data.shape = (2, 3)
        >>> g['a'] = BaseType(name='a', data=data, shape=data.shape, type=Float32, dimensions=('x', 'y'))
        >>> g['x'] = BaseType(name='x', data=numpy.arange(2.), shape=(2,), type=Float64)
        >>> g['y'] = BaseType(name='y', data=numpy.arange(3.), shape=(3,), type=Float64)
        >>> print g.array.data
        [[ 0.  1.  2.]
         [ 3.  4.  5.]]
    
    We can treat the Grid like an array and slice it. This will return a new
    grid object with the proper data and axes::

        >>> print g[:,0]
        <class 'pydap.model.GridType'>
            with data
        [[ 0.]
         [ 3.]]
            and axes
        [ 0.  1.]
        [ 0.]

    We can also use a shortcut notation for maps and the array::

        >>> print g['y'][0]
        0.0
        >>> print g.y[:]
        [ 0.  1.  2.]

    A nice thing about Grids is that we can slice them by the maps::

        >>> print g[ :,(g.y > 0) ]
        <class 'pydap.model.GridType'>
            with data
        [[ 1.  2.]
         [ 4.  5.]]
            and axes
        [ 0.  1.]
        [ 1.  2.]

    Though for remote Grids (ie, on Opendap servers) this only works
    for continuous conditions, ie, with a start and an end index.

    """
    def __str__(self):
        return """%s
    with data
%s
    and axes
%s""" % (self.__class__, self.data[0], '\n'.join(map(str, self.data[1:])))

    def __getitem__(self, key):
        if isinstance(key, basestring):
            return StructureType.__getitem__(self, key)
        else:
            key = fix_slice(key, self.shape)
            out = copy.deepcopy(self)
            for var, slice_ in zip(out.walk(), [key] + list(key)):
                var.data = var.data[slice_]
                var.shape = var.data.shape
            return out

    @property
    def __array_interface__(self):
        data = numpy.asarray(self.array.data)
        return {
                'version': 3,
                'shape': data.shape,
                'typestr': data.dtype.str,
                'data': data,
                }

    @property
    def array(self):
        return self[self._keys[0]]

    @property
    def maps(self):
        return odict((k, self[k]) for k in self._keys[1:])

    @property
    def dimensions(self):
        return tuple(self._keys[1:])

    # These attributes all come from the array object.
    def __len__(self): return len(self.array)
    @property
    def type(self): return self.array.type
    def _get_shape(self): return self.array.shape
    def _set_shape(self, shape): self.array.shape = shape
    shape = property(_get_shape, _set_shape)
    


def _test():
    import doctest
    doctest.testmod()


if __name__ == "__main__":
    _test()



import re
from urlparse import urlsplit, urlunsplit
import logging

import httplib2
from pydap.util import socks
httplib2.socks = socks

import pydap.lib
from pydap.exceptions import ServerError


log = logging.getLogger('pydap')


def request(url):
    """
    Open a given URL and return headers and body.

    This function retrieves data from a given URL, returning the headers
    and the response body. Authentication can be set by adding the
    username and password to the URL; this will be sent as clear text
    only if the server only supports Basic authentication.

    """
    h = httplib2.Http(cache=pydap.lib.CACHE,
            timeout=pydap.lib.TIMEOUT,
            proxy_info=pydap.lib.PROXY)
    scheme, netloc, path, query, fragment = urlsplit(url)
    if '@' in netloc:
        credentials, netloc = netloc.split('@', 1)  # remove credentials from netloc
        username, password = credentials.split(':', 1)
        h.add_credentials(username, password)

    url = urlunsplit((
            scheme, netloc, path, query, fragment
            )).rstrip('?&')

    log.info('Opening %s' % url)
    resp, data = h.request(url, "GET",
            headers = {'user-agent': pydap.lib.USER_AGENT})

    # When an error is returned, we parse the error message from the
    # server and return it in a ``ClientError`` exception.
    if resp.get("content-description") in ["dods_error", "dods-error"]:
        m = re.search('code = (?P<code>[^;]+);\s*message = "(?P<msg>.*)"',
                data, re.DOTALL | re.MULTILINE)
        msg = 'Server error %(code)s: "%(msg)s"' % m.groupdict()
        raise ServerError(msg)

    return resp, data
